import requests
import os

from django import forms
from django.conf import settings
from django.core.exceptions import ValidationError

from tom_tns.tns_api import (get_tns_values, group_names, get_reverse_tns_values,
                             pre_upload_files_to_tns, submit_through_hermes)
from tom_dataproducts.models import DataProduct

from crispy_forms.helper import FormHelper
from crispy_forms.layout import Layout, Row, Column, Submit, HTML
from crispy_forms.bootstrap import AppendedText, Alert, Accordion, AccordionGroup
from datetime import datetime


HERMES_FLUX_UNITS = [
    ("AB mag", "AB mag"), ("Vega mag", "Vega mag"),
    ("mJy", "mJy"), ("erg / s / cm² / Å", "erg / s / cm² / Å")
]


class BaseReportForm(forms.Form):
    def is_set(self, field):
        if field in self.cleaned_data and self.cleaned_data[field]:
            return True
        return False


class TNSReportForm(BaseReportForm):
    submitter = forms.CharField(required=True, widget=forms.HiddenInput())
    object_name = forms.CharField(required=True, widget=forms.HiddenInput())
    ra = forms.FloatField(label='R.A.')
    dec = forms.FloatField(label='Dec.')
    reporting_group = forms.ChoiceField(choices=[])
    discovery_data_source = forms.ChoiceField(choices=[])
    reporter = forms.CharField(widget=forms.Textarea(attrs={'rows': 1}), label='Reporter Name(s) / Author List')
    discovery_date = forms.DateTimeField(initial=datetime.utcnow())
    at_type = forms.ChoiceField(choices=[], label='AT type')
    archive = forms.ChoiceField(required=False, choices=[])
    archival_remarks = forms.CharField(required=False)
    nondetection_observation_date = forms.DateTimeField(required=False, label='Observation date')
    nondetection_flux = forms.FloatField(required=False, label='Flux')
    nondetection_flux_units = forms.ChoiceField(choices=[], required=False, label='Flux units')
    nondetection_filter = forms.ChoiceField(choices=[], required=False, label='Filter')
    nondetection_instrument = forms.ChoiceField(choices=[], required=False, label='Instrument')
    nondetection_observer = forms.CharField(required=False, label='Observer')
    nondetection_exposure_time = forms.FloatField(required=False, label='Exposure time')
    observation_date = forms.DateTimeField()
    flux = forms.FloatField()
    flux_error = forms.FloatField()
    flux_units = forms.ChoiceField(choices=[])
    filter = forms.ChoiceField(choices=[])
    instrument = forms.ChoiceField(choices=[])
    limiting_flux = forms.FloatField(required=False)
    exposure_time = forms.FloatField(required=False)
    observer = forms.CharField(required=False)
    discovery_remarks = forms.CharField(required=False, widget=forms.Textarea(attrs={'rows': 2}))
    photometry_remarks = forms.CharField(required=False, widget=forms.Textarea(attrs={'rows': 2}))
    nondetection_remarks = forms.CharField(required=False, widget=forms.Textarea(attrs={'rows': 2}))

    def __init__(self, *args, **kwargs):
        """
        Initialize choices from TNS API values, and set common defaults.
        Also define the form layout using crispy_forms.
        """
        super().__init__(*args, **kwargs)
        self.fields['discovery_data_source'].choices = sorted(get_tns_values('groups'), key=lambda x: x[1])
        self.fields['at_type'].choices = get_tns_values('at_types')
        self.fields['at_type'].initial = (1, "PSN")
        self.fields['filter'].choices = get_tns_values('filters')
        self.fields['filter'].initial = ("22", "r-Sloan")
        self.fields['nondetection_filter'].choices = get_tns_values('filters')
        self.fields['nondetection_filter'].initial = ("22", "r-Sloan")
        self.fields['archive'].choices = get_tns_values('archives')
        self.fields['archive'].initial = ("0", "Other")
        self.fields['instrument'].choices = get_tns_values('instruments')
        self.fields['instrument'].initial = ("0", "Other")
        self.fields['nondetection_instrument'].choices = get_tns_values('instruments')
        self.fields['nondetection_instrument'].initial = ("0", "Other")
        if submit_through_hermes():
            self.fields['flux_units'].choices = HERMES_FLUX_UNITS
            self.fields['flux_units'].initial = ("AB mag", "AB mag")
            self.fields['nondetection_flux_units'].choices = HERMES_FLUX_UNITS
            self.fields['nondetection_flux_units'].initial = ("AB mag", "AB mag")
        else:
            self.fields['flux_units'].choices = get_tns_values('units')
            self.fields['flux_units'].initial = (1, "ABMag")
            self.fields['nondetection_flux_units'].choices = get_tns_values('units')
            self.fields['nondetection_flux_units'].initial = (1, "ABMag")

        # set choices of reporting groups to list set in settings.py
        bot_tns_group_names = group_names()
        if not bot_tns_group_names:
            bot_tns_group_names = [settings.TOM_NAME]
        tns_group_list = []
        for bot_group_name in bot_tns_group_names:
            if get_reverse_tns_values('groups', bot_group_name):
                tns_group_list.append(get_reverse_tns_values('groups', bot_group_name))
        tns_group_list.append(get_reverse_tns_values('groups', 'None'))
        self.fields['reporting_group'].choices = tns_group_list
        # set initial group for discovery source if tom_name is in the list of tns group names
        default_group = tns_group_list[0]
        if default_group:
            self.fields['discovery_data_source'].initial = default_group

        self.helper = FormHelper()
        self.helper.layout = Layout(
            'submitter', 'object_name',
            Row(
                Column('reporter', css_class='col-md-6'),
                Column('reporting_group'),
                Column('discovery_data_source'),
            ),
            Row(
                Column(AppendedText('ra', 'deg')),
                Column(AppendedText('dec', 'deg')),
                Column('discovery_date'),
                Column('at_type'),
            ),
            Row(Column('discovery_remarks')),
            Row(HTML('<h4>Discovery Photometry</h4>')),
            Row(
                Column('observation_date'),
                Column('instrument'),
                Column('limiting_flux'),
                Column('observer'),
            ),
            Row(
                Column('flux'),
                Column('flux_error'),
                Column('flux_units'),
                Column('filter'),
            ),
            Row(Column('photometry_remarks')),
            Row(HTML('<h4>Last Nondetection</h4>')),
            Alert(
                content="""Fill out either the Archive information,
                           or the Detection information for the last nondetection
                        """,
                css_class='alert-warning'
            ),
            Accordion(
                AccordionGroup('Archive information',
                               Row(
                                    Column('archive'),
                                    Column('archival_remarks'),
                                  ),
                               ),
                AccordionGroup('Detection information',
                               Row(
                                    Column('nondetection_observation_date'),
                                    Column('nondetection_instrument'),
                                    Column('nondetection_exposure_time'),
                                    Column('nondetection_observer'),
                                  ),
                               Row(
                                    Column('nondetection_flux'),
                                    Column('nondetection_flux_units'),
                                    Column('nondetection_filter'),
                                  ),
                               Row(Column('nondetection_remarks', css_class='col-md-12'))
                               )
            ),
            Row(Column(Submit('submit', 'Submit Report'))),
        )

    def clean(self):
        super().clean()
        # Either the archival info or nondetection flux info must be set for a valid TNS submission
        if not self.is_set('archive') or not self.is_set('archival_remarks'):
            if any([not self.is_set(field) for field in [
                'nondetection_flux', 'nondetection_instrument', 'nondetection_filter', 'nondetection_observation_date'
            ]]):
                raise ValidationError(
                    "Must set either last nondetection archival information,"
                    "or last nondetection flux, obsdate, filter and instrument"
                )

    def generate_hermes_report(self):
        """
        Generate Hermes TNS discovery report according to the hermes schema

        Returns the report as a Dict to be sent as JSON
        """
        hermes_report = {
            'topic': 'hermes.test',
            'title': f'{self.cleaned_data["object_name"]} TNS discovery report',
            'submit_to_tns': True,
            'submitter': self.cleaned_data['submitter'],
            'authors': self.cleaned_data['reporter'],  # TODO: Set the initial authors on the field from the settings
            'data': {
                'targets': [{
                    'name': self.cleaned_data['object_name'],
                    'ra': self.cleaned_data['ra'],
                    'dec': self.cleaned_data['dec'],
                    'new_discovery': True,
                    'discovery_info': {
                        'date': self.cleaned_data['discovery_date'].isoformat(),
                        'discovery_source': dict(
                            self.fields['discovery_data_source'].choices)[self.cleaned_data['discovery_data_source']],
                        'reporting_group': dict(
                            self.fields['reporting_group'].choices)[self.cleaned_data['reporting_group']],
                        'transient_type': dict(self.fields['at_type'].choices)[int(self.cleaned_data['at_type'])],
                    }
                }],
                'photometry': [{
                    'target_name': self.cleaned_data['object_name'],
                    'date_obs': self.cleaned_data['observation_date'].isoformat(),
                    'instrument': dict(self.fields['instrument'].choices)[self.cleaned_data['instrument']],
                    'bandpass': dict(self.fields['filter'].choices)[self.cleaned_data['filter']],
                    "brightness": self.cleaned_data['flux'],
                    "brightness_error": self.cleaned_data['flux_error'],
                    "brightness_unit": self.cleaned_data['flux_units'],
                }]
            },
        }
        if self.is_set('discovery_remarks'):
            hermes_report['data']['targets'][0]['comments'] = self.cleaned_data['discovery_remarks']
        if self.is_set('photometry_remarks'):
            hermes_report['data']['photometry'][0]['comments'] = self.cleaned_data['photometry_remarks']
        if self.is_set('exposure_time'):
            hermes_report['data']['photometry'][0]['exposure_time'] = self.cleaned_data['exposure_time']
        if self.is_set('observer'):
            hermes_report['data']['photometry'][0]['observer'] = self.cleaned_data['observer']
        if self.is_set('limiting_flux'):
            hermes_report['data']['photometry'][0]['limiting_brightness'] = self.cleaned_data['limiting_flux']

        if self.is_set('archive') and self.is_set('archival_remarks'):
            discovery_info = hermes_report['data']['targets'][0]['discovery_info']
            discovery_info['nondetection_source'] = dict(
                self.fields['archive'].choices)[self.cleaned_data['archive']]
            discovery_info['nondetection_comments'] = self.cleaned_data['archival_remarks']
        elif all([self.is_set(field) for field in [
            'nondetection_flux', 'nondetection_instrument', 'nondetection_filter', 'nondetection_observation_date'
        ]]):
            nondetection = {
                'limiting_brightness': self.cleaned_data['nondetection_flux'],
                'limiting_brightness_unit': self.cleaned_data['nondetection_flux_units'],
                'date_obs': self.cleaned_data['nondetection_observation_date'],
                'bandpass': self.cleaned_data['nondetection_filter'],
                'instrument': self.cleaned_data['nondetection_instrument'],
            }
            if self.is_set('nondetection_exposure_time'):
                nondetection['exposure_time'] = self.cleaned_data['nondetection_exposure_time']
            if self.is_set('nondetection_observer'):
                nondetection['observer'] = self.cleaned_data['nondetection_observer']
            if self.is_set('nondetection_remarks'):
                nondetection['comments'] = self.cleaned_data['nondetection_remarks']
            hermes_report['data']['photometry'].append(nondetection)

        return hermes_report, []

    def generate_tns_report(self):
        """
        Generate TNS bulk transient report according to the schema in this manual:
        https://sandbox.wis-tns.org/sites/default/files/api/TNS_bulk_reports_manual.pdf

        Returns the report as a JSON-formatted string
        """
        report_data = {
            "at_report": {
                "0": {
                    "ra": {
                        "value": self.cleaned_data['ra'],
                    },
                    "dec": {
                        "value": self.cleaned_data['dec'],
                    },
                    "reporting_group_id": self.cleaned_data['reporting_group'],
                    "discovery_data_source_id": self.cleaned_data['discovery_data_source'],
                    "reporter": self.cleaned_data['reporter'],
                    "discovery_datetime": self.cleaned_data['discovery_date'].strftime('%Y-%m-%d %H:%M:%S'),
                    "at_type": self.cleaned_data['at_type'],
                    "remarks": self.cleaned_data["discovery_remarks"],
                    "non_detection": {
                        "archiveid": self.cleaned_data['archive'],
                        "archival_remarks": self.cleaned_data['archival_remarks'],
                        "comments": self.cleaned_data['nondetection_remarks'],
                    },
                    "photometry": {
                        "photometry_group": {
                            "0": {
                                "obsdate": self.cleaned_data['observation_date'].strftime('%Y-%m-%d %H:%M:%S'),
                                "flux": self.cleaned_data['flux'],
                                "flux_error": self.cleaned_data['flux_error'],
                                "flux_units": self.cleaned_data['flux_units'],
                                "filter_value": self.cleaned_data['filter'],
                                "instrument_value": self.cleaned_data['instrument'],
                                "limiting_flux": self.cleaned_data['limiting_flux'],
                                "exptime": self.cleaned_data['exposure_time'],
                                "observer": self.cleaned_data['observer'],
                                "comments": self.cleaned_data['photometry_remarks'],
                            },
                        }
                    },
                }
            }
        }
        return report_data


class TNSClassifyForm(BaseReportForm):
    submitter = forms.CharField(required=False, widget=forms.HiddenInput())
    object_name = forms.CharField(required=False)
    ra = forms.FloatField(required=False, widget=forms.HiddenInput())
    dec = forms.FloatField(required=False, widget=forms.HiddenInput())
    classifier = forms.CharField(widget=forms.Textarea(attrs={'rows': 1}), label='Classifier Name(s) / Author List')
    classification = forms.ChoiceField(choices=[])
    redshift = forms.FloatField(required=False, widget=forms.TextInput())
    reporting_group = forms.ChoiceField(choices=[])
    classification_remarks = forms.CharField(required=False, widget=forms.Textarea(attrs={'rows': 2}))
    observation_date = forms.DateTimeField()
    instrument = forms.ChoiceField(choices=[])
    exposure_time = forms.FloatField(required=False)
    observer = forms.CharField()
    reducer = forms.CharField(required=False)
    spectrum_type = forms.ChoiceField(choices=[])
    ascii_file = forms.ChoiceField(label='ASCII file', choices=[], required=True)
    fits_file = forms.ChoiceField(label='FITS file', choices=[], required=False)
    ascii_file_description = forms.CharField(label='ASCII file description', required=False)
    fits_file_description = forms.CharField(label='FITS file description', required=False)
    ascii_file_override = forms.FileField(label='ASCII file override', required=False, widget=forms.FileInput())
    fits_file_override = forms.FileField(label='FITS file override', required=False, widget=forms.FileInput())
    spectrum_remarks = forms.CharField(required=False, widget=forms.Textarea(attrs={'rows': 2}))

    def __init__(self, *args, **kwargs):
        """
        Set initial choices for the classification form using the TNS API values and set default values.
        Also define the form layout using crispy-forms.
        """
        super().__init__(*args, **kwargs)
        self.fields['instrument'].choices = get_tns_values('instruments')
        self.fields['instrument'].initial = (0, "Other")
        self.fields['classification'].choices = get_tns_values('object_types')
        self.fields['classification'].initial = (1, "SN")
        self.fields['spectrum_type'].choices = get_tns_values('spectra_types')
        self.fields['ascii_file'].choices = kwargs['initial']['ascii_file_choices']
        self.fields['fits_file'].choices = kwargs['initial']['fits_file_choices']

        # set choices of reporting groups to list set in settings.py
        bot_tns_group_names = group_names()
        if not bot_tns_group_names:
            bot_tns_group_names = [settings.TOM_NAME]
        tns_group_list = []
        for bot_group_name in bot_tns_group_names:
            if get_reverse_tns_values('groups', bot_group_name):
                tns_group_list.append(get_reverse_tns_values('groups', bot_group_name))
        tns_group_list.append(get_reverse_tns_values('groups', 'None'))
        self.fields['reporting_group'].choices = tns_group_list

        self.helper = FormHelper()
        self.helper.layout = Layout(
            'submitter', 'ra', 'dec',
            Row(
                Column('classifier', css_class='col-md-8'),
                Column('reporting_group'),
            ),
            Row(
                Column('object_name'),
                Column('classification'),
                Column('redshift'),
            ),
            Row(Column('classification_remarks')),
            Row(HTML('<h4>Classification Spectrum</h4>')),
            Row(
                Column('observation_date'),
                Column('observer'),
                Column('reducer'),
            ),
            Row(
                Column('instrument'),
                Column('exposure_time'),
                Column('spectrum_type'),
            ),
            Row(
                Column('ascii_file'),
                Column('fits_file'),
            ),
            Row(
                Column('ascii_file_override'),
                Column('fits_file_override'),
            ),
            Row(
                Column('ascii_file_description'),
                Column('fits_file_description'),
            ),
            Row(Column('spectrum_remarks')),
            Row(Column(Submit('submit', 'Submit Classification'))),
        )

    def generate_hermes_report(self):
        """
        Generate Hermes TNS classification report according to the hermes schema

        Returns the report as a Dict to be sent as JSON
        """
        if self.is_set('ascii_file_override'):
            ascii_file = self.cleaned_data['ascii_file_override']
        else:
            ascii_file = DataProduct.objects.get(pk=self.cleaned_data['ascii_file']).data
        if self.is_set('fits_file_override'):
            fits_file = self.cleaned_data['fits_file_override']
        elif self.is_set('fits_file'):
            fits_file = DataProduct.objects.get(pk=self.cleaned_data['fits_file']).data
        else:
            fits_file = None
        hermes_report = {
            'topic': 'hermes.test',
            'title': f'{self.cleaned_data["object_name"]} TNS classification report',
            'submit_to_tns': True,
            'submitter': self.cleaned_data['submitter'],
            'authors': self.cleaned_data['classifier'],  # TODO: Set the initial authors on the field from the settings
            'data': {
                'targets': [{
                    'name': self.cleaned_data['object_name'],
                    'ra': self.cleaned_data['ra'],
                    'dec': self.cleaned_data['dec'],
                    'new_discovery': False,
                    'discovery_info': {
                        'reporting_group': dict(
                            self.fields['reporting_group'].choices)[self.cleaned_data['reporting_group']]
                    }
                }],
                'spectroscopy': [{
                    'target_name': self.cleaned_data['object_name'],
                    'date_obs': self.cleaned_data['observation_date'].isoformat(),
                    'instrument': dict(self.fields['instrument'].choices)[self.cleaned_data['instrument']],
                    'spec_type': dict(self.fields['spectrum_type'].choices)[self.cleaned_data['spectrum_type']],
                    'classification': dict(self.fields['classification'].choices)[self.cleaned_data['classification']],
                    'observer': self.cleaned_data['observer'],
                    'file_info': [
                        {
                            'name': os.path.basename(ascii_file.name),
                            'description': self.cleaned_data.get('ascii_file_description', '')
                        }
                    ]
                }]
            },
        }
        files = [ascii_file]
        if self.is_set('redshift'):
            hermes_report['data']['targets'][0]['redshift'] = self.cleaned_data['redshift']
        if self.is_set('classification_remarks'):
            hermes_report['data']['targets'][0]['comments'] = self.cleaned_data['classification_remarks']
        if self.is_set('reducer'):
            hermes_report['data']['spectroscopy'][0]['reducer'] = self.cleaned_data['reducer']
        if self.is_set('exposure_time'):
            hermes_report['data']['spectroscopy'][0]['exposure_time'] = self.cleaned_data['exposure_time']
        if fits_file:
            hermes_report['data']['spectroscopy'][0]['file_info'].append({
                'name': os.path.basename(fits_file.name),
                'description': self.cleaned_data.get('fits_file_description', '')
            })
            files.append(fits_file)
        if self.is_set('spectrum_remarks'):
            hermes_report['data']['spectroscopy'][0]['comments'] = self.cleaned_data['spectrum_remarks']

        return hermes_report, files

    def generate_tns_report(self):
        """
        Generate TNS bulk classification report according to the schema in this manual:
        https://sandbox.wis-tns.org/sites/default/files/api/TNS_bulk_reports_manual.pdf

        Returns the report as a JSON-formatted string
        """
        if self.is_set('ascii_file_override'):
            ascii_file = self.cleaned_data['ascii_file_override']
        else:
            ascii_file = DataProduct.objects.get(pk=self.cleaned_data['ascii_file']).data
        if self.is_set('fits_file_override'):
            fits_file = self.cleaned_data['fits_file_override']
        elif self.is_set('fits_file'):
            fits_file = DataProduct.objects.get(pk=self.cleaned_data['fits_file']).data
        else:
            fits_file = None
        file_list = {'ascii_file': ascii_file,
                     'fits_file': fits_file,
                     'other_files': []}
        try:
            tns_filenames = pre_upload_files_to_tns(file_list)
        except requests.exceptions.HTTPError as e:
            return {'message': f"ERROR: {e}"}
        report_data = {
            "classification_report": {
                "0": {
                    "name": self.cleaned_data['object_name'],
                    "classifier": self.cleaned_data['classifier'],
                    "objtypeid": self.cleaned_data['classification'],
                    "redshift": self.cleaned_data['redshift'],
                    "groupid": self.cleaned_data['reporting_group'],
                    "remarks": self.cleaned_data['classification_remarks'],
                    "spectra": {
                        "spectra-group": {
                            "0": {
                                "obsdate": self.cleaned_data['observation_date'].strftime('%Y-%m-%d %H:%M:%S'),
                                "instrumentid": self.cleaned_data['instrument'],
                                "exptime": self.cleaned_data['exposure_time'],
                                "observer": self.cleaned_data['observer'],
                                "reducer": self.cleaned_data['reducer'],
                                "spectypeid": self.cleaned_data['spectrum_type'],
                                "ascii_file": tns_filenames.get('ascii_file', ''),
                                "fits_file": tns_filenames.get('fits_file', ''),
                                "remarks": self.cleaned_data['spectrum_remarks'],
                            },
                        }
                    },
                }
            }
        }
        return report_data



import requests
from urllib.parse import urljoin

from django.core.cache import cache
from django.conf import settings
from django.contrib import messages

import json
import time
import logging
logger = logging.getLogger(__name__)


class BadTnsRequest(Exception):
    """ This Exception will be raised by errors during the TNS submission process """
    pass


def submit_through_hermes():
    """ Check if hermes credentials exist and are setup so TNS messages should be sent through hermes
        Returns True if it should go through hermes, False if it should go directly to TNS
    """
    if hasattr(settings, 'DATA_SHARING') and settings.DATA_SHARING.get('hermes', {}).get('ENABLE_TNS', False):
        return True
    return False


def map_filter_to_tns(filter):
    """ Checks if a filter mapping was set in the settings, and if so returns the mapped value for the filter passed in
    """
    if submit_through_hermes():
        return settings.DATA_SHARING.get('hermes', {}).get('FILTER_MAPPING', {}).get(filter)
    else:
        return settings.BROKERS.get('TNS', {}).get('filter_mapping', {}).get(filter)


def default_authors():
    """ Returns default authors if set in the settings, otherwise empty string.
    """
    if submit_through_hermes():
        return settings.DATA_SHARING.get('hermes', {}).get('DEFAULT_AUTHORS', '')
    else:
        return settings.BROKERS.get('TNS', {}).get('default_authors', '')


def group_names():
    """ Returns the Group names list from settings, either from Hermes config or TNS config
    """
    if submit_through_hermes():
        return settings.DATA_SHARING.get('hermes', {}).get('GROUP_NAMES', [])
    else:
        return settings.BROKERS.get('TNS', {}).get('group_names', [])


def get_tns_credentials():
    """
    Get the TNS credentials from settings.py.
    This should include the bot_id, bot_name, api_key, tns_base_url, and possibly group_name.
    """
    try:
        tns_info = settings.BROKERS['TNS']

        # Build TNS Marker using Bot info if API key is present
        if tns_info.get('api_key', None):
            tns_info['marker'] = 'tns_marker' + json.dumps({'tns_id': tns_info.get('bot_id', None),
                                                            'type': 'bot',
                                                            'name': tns_info.get('bot_name', None)})
        else:
            logger.error("TNS API key not found in settings.py")
            tns_info = {}
    except (KeyError, AttributeError):
        logger.error("TNS credentials not found in settings.py")
        tns_info = {}
    return tns_info


def get_tns_values(option_list):
    """ Retrieve the TNS options. These are cached for one hour.
    Returns a list of tuples, each tuple containing the option value and the option label.
    """
    all_tns_values = cache.get("all_tns_values", {})
    if not all_tns_values:
        all_tns_values, _ = populate_tns_values()
    selected_values = all_tns_values.get(option_list, [])
    tuple_list = []
    if isinstance(selected_values, list):
        tuple_list = [(i, v) for i, v in enumerate(selected_values)]
    if isinstance(selected_values, dict):
        tuple_list = [(k, v) for k, v in selected_values.items()]
    return tuple_list


def get_reverse_tns_values(option_list, value):
    """
    Retrieve the reverse mapping of TNS options. Used to go from option to value for a specific list of options.
    Returns a tuple of the option value and the option label.
    """
    reversed_tns_values = cache.get("reverse_tns_values", {})
    if not reversed_tns_values:
        _, reversed_tns_values = populate_tns_values()
    try:
        return reversed_tns_values[option_list][value], value
    except KeyError:
        return None


def populate_tns_values():
    """pull all the values from the TNS API and Cache for an hour"""
    all_tns_values = {}
    reversed_tns_values = {}
    if submit_through_hermes():
        # Get the tns values from the HERMES api
        hermes_tns_options_url = urljoin(settings.DATA_SHARING.get('hermes', {}).get(
            'BASE_URL', ''), 'api/v0/tns_options/')
        headers = {'Authorization': f"Token {settings.DATA_SHARING.get('hermes', {}).get('HERMES_API_KEY', '')}"}
        try:
            resp = requests.get(hermes_tns_options_url, headers=headers)
            resp.raise_for_status()
            all_tns_values = resp.json()
        except Exception as e:
            logging.warning(f"Failed to retrieve tns values from Hermes: {repr(e)}")
    else:
        # Need to spoof a web based user agent or TNS will block the request :(
        SPOOF_USER_AGENT = 'Mozilla/5.0 (X11; Linux i686; rv:110.0) Gecko/20100101 Firefox/110.0.'

        # Use sandbox URL if no url found in settings.py
        tns_base_url = get_tns_credentials().get('tns_base_url', 'https://sandbox.wis-tns.org/api')
        try:
            resp = requests.get(urljoin(tns_base_url, 'api/values/'),
                                headers={'user-agent': SPOOF_USER_AGENT})
            resp.raise_for_status()
            all_tns_values = resp.json().get('data', {})

        except Exception as e:
            logging.warning(f"Failed to retrieve tns values: {repr(e)}")

    if all_tns_values:
        reversed_tns_values = reverse_tns_values(all_tns_values)
        cache.set("all_tns_values", all_tns_values, 3600)
        cache.set("reverse_tns_values", reversed_tns_values, 3600)
    return all_tns_values, reversed_tns_values


def reverse_tns_values(all_tns_values):
    """reverse the values from the TNS API"""
    reversed_tns_values = {}
    for key, values in all_tns_values.items():
        if isinstance(values, list):
            reversed_tns_values[key] = {value: index for index, value in enumerate(values)}
        elif isinstance(values, dict):
            reversed_tns_values[key] = {v: k for k, v in values.items()}
    return reversed_tns_values


def build_file_dict(files):
    """
    Build a dictionary of files to upload to the TNS as well as a dictionary connecting the uploaded name to the new
    name returned from TNS.
    TNS requires a specific format for the file upload:
    https://www.wis-tns.org/sites/default/files/api/TNS_bulk_reports_manual.pdf
    file_Load: {files[0]: Filename, files[1]: Filename2, ...}
    new_files: {ascii_file: <<index for ascii file>>, fits_file: <<index for fits file>>, ...}
    """
    new_files = {}
    file_load = {}
    i = 0
    if files['ascii_file']:
        file_load[f'files[{i}]'] = (files['ascii_file'].name, files['ascii_file'].open(), 'text/plain')
        new_files['ascii_file'] = i
        i += 1
    if files['fits_file']:
        file_load[f'files[{i}]'] = (files['fits_file'].name, files['fits_file'].open('rb'), 'application/fits')
        new_files['fits_file'] = i
        i += 1
    return file_load, new_files


def pre_upload_files_to_tns(files):
    """
    Upload files to the Transient Name Server according to this manual:
    https://sandbox.wis-tns.org/sites/default/files/api/TNS_bulk_reports_manual.pdf
    """
    tns_credentials = get_tns_credentials()
    file_load, new_files = build_file_dict(files)
    if not file_load:
        return None
    # build request parameters
    tns_marker = tns_credentials['marker']
    upload_data = {'api_key': tns_credentials['api_key']}
    response = requests.post(tns_credentials['tns_base_url'] + '/file-upload', headers={'User-Agent': tns_marker},
                             data=upload_data, files=file_load)
    response.raise_for_status()
    # If successful, TNS returns a list of new filenames
    new_filenames = response.json().get('data', {})
    logger.info(f"Uploaded {', '.join(new_filenames)} to the TNS")
    if not new_filenames:
        return None
    # Return dictionary of updated TNS names for each uploaded file_type
    for file in new_files:
        try:
            new_files[file] = new_filenames[new_files[file]]
        except IndexError:
            new_files[file] = ''
    return new_files


def send_tns_report(data):
    """
    Send a JSON bulk report to the Transient Name Server according to this manual:
    https://sandbox.wis-tns.org/sites/default/files/api/TNS_bulk_reports_manual.pdf
    Returns a report ID if successful.
    This ID can be used to retrieve the report from the TNS.
    """
    tns_info = get_tns_credentials()
    json_data = {'api_key': tns_info['api_key'], 'data': data}
    response = requests.post(tns_info['tns_base_url'] + '/bulk-report',
                             headers={'User-Agent': tns_info['marker']},
                             data=json_data)
    response.raise_for_status()
    report_id = response.json()['data']['report_id']
    logger.info(f'Sent TNS report ID {report_id:d}')
    return report_id


def parse_object_from_tns_response(response_json, request):
    feedback_section = response_json['data']['feedback']
    feedbacks = []
    iau_name = None
    if 'at_report' in feedback_section:
        feedbacks += feedback_section['at_report']
    if 'classification_report' in feedback_section:
        feedbacks += feedback_section['classification_report'][0]['classification_messages']
    for feedback in feedbacks:
        if '100' in feedback:  # transient object was inserted
            iau_name = 'AT' + feedback['100']['objname']
            log_message = f'New transient {iau_name} was created'
            logger.info(log_message)
            messages.success(request, log_message)
            break
        elif '101' in feedback:  # transient object exists
            iau_name = feedback['101']['prefix'] + feedback['101']['objname']
            log_message = f'Existing transient {iau_name} was reported'
            logger.info(log_message)
            messages.info(request, log_message)
            break
        elif '121' in feedback:  # object name prefix has changed
            iau_name = feedback['121']['new_object_name']
            log_message = f'Transient name changed to {iau_name}'
            logger.info(log_message)
            messages.success(request, log_message)
            break
    else:  # If neither 'classification_report', nor 'at_report' were in the feedback section
        log_message = 'No recognized feedback in the TNS response.'
        logger.error(log_message)
        messages.error(request, log_message)
    return iau_name


def get_tns_report_reply(report_id, request):
    """
    Get feedback from the Transient Name Server in response to a bulk report according to this manual:
    https://sandbox.wis-tns.org/sites/default/files/api/TNS_bulk_reports_manual.pdf

    Posts an informational message in a banner on the page using ``request``
    """
    tns_info = get_tns_credentials()
    reply_data = {'api_key': tns_info['api_key'], 'report_id': report_id}
    iau_name = None
    attempts = 0
    # TNS Submissions return immediately with an id, which you must then check to see if the message
    # was processed, and if it was accepted or rejected. Here we check up to 10 times, waiting 1s
    # between checks. Under normal circumstances, it should be processed within a few seconds.
    while attempts < 10:
        response = requests.post(tns_info['tns_base_url'] + '/bulk-report-reply',
                                 headers={'User-Agent': tns_info['marker']}, data=reply_data)
        attempts += 1
        # A 404 response means the report has not been processed yet
        if response.status_code == 404:
            time.sleep(1)
        # A 400 response means the report failed with certain errors
        elif response.status_code == 400:
            raise BadTnsRequest(f"TNS submission failed with feedback: "
                                f"{response.json().get('data', {}).get('feedback', {})}")
        # A 200 response means the report was successful, and we can parse out the object name
        elif response.status_code == 200:
            iau_name = parse_object_from_tns_response(response.json(), request)
            break
        else:
            raise BadTnsRequest(f"TNS submission failed with status code {response.status_code}")
    if not iau_name:
        raise BadTnsRequest(f"TNS submission failed to be processed within 10 seconds. The report_id = {report_id}")
    return iau_name



from django.apps import AppConfig
from django.urls import path, include


class TomTnsConfig(AppConfig):
    default_auto_field = 'django.db.models.BigAutoField'
    name = 'tom_tns'

    def target_detail_buttons(self):
        """
        Integration point for adding buttons to the target detail view.
        This method should return a list of dictionaries, each containing the keys:
        - 'namespace': The namespace of the app that provides the button's view
        - 'title': The title of the button
        - 'class': The CSS class of the button
        - 'text': The text of the button
        """
        return {'namespace': 'tns:report-tns',
                'title': 'TNS',
                'class': 'btn  btn-info',
                'text': 'TNS',
                }

    def include_url_paths(self):
        """
        Integration point for adding URL patterns to the Tom Common URL configuration.
        This method should return a list of URL patterns to be included in the main URL configuration.
        """
        urlpatterns = [
            path('tns/', include('tom_tns.urls', namespace='tns'))
        ]
        return urlpatterns



from django.conf import settings
from django.contrib import messages
from urllib.parse import urljoin

import requests
import json
import logging
import os

logger = logging.getLogger(__name__)


def get_object_from_response(response_json):
    # If the submission is successful, then pull out the tns_object name from the references and return it
    references = response_json.get('data', {}).get('references', [])
    for reference in references:
        if reference.get('source') == 'tns_object':
            return reference.get('citation')
    return None


def submit_to_hermes(hermes_message, files, request):
    hermes_submit_url = urljoin(settings.DATA_SHARING.get('hermes', {}).get('BASE_URL', ''), 'api/v0/submit_message/')
    headers = {'Authorization': f"Token {settings.DATA_SHARING.get('hermes', {}).get('HERMES_API_KEY', '')}"}
    response_json = {}
    try:
        if not files:
            # Can submit simple json payload to hermes, and this assumed to be a new discovery
            response = requests.post(url=hermes_submit_url, json=hermes_message, headers=headers)
            response_json = response.json()
            response.raise_for_status()
            logger.info(f"Sent TNS discovery message through Hermes with uuid {response_json.get('uuid')}")
            return get_object_from_response(response_json)
        else:
            # There are files, so this must be a classification submission
            data = {'data': json.dumps(hermes_message)}
            files_to_submit = []
            for file in files:
                content_type = 'text/plain'
                if file.name.endswith('fits') or file.name.endswith('fits.fz'):
                    content_type = 'application/fits'
                files_to_submit.append(('files', (os.path.basename(file.name), file.open('rb'), content_type)))
            response = requests.post(url=hermes_submit_url, data=data, files=files_to_submit, headers=headers)
            response_json = response.json()
            response.raise_for_status()
            logger.info(f"Sent TNS classification message through Hermes with uuid {response_json.get('uuid')}")
            return get_object_from_response(response_json)
    except Exception as e:
        error_msg = 'Failed to Submit message to Hermes/TNS'
        if response_json:
            error_msg += f': {response_json}'
        else:
            error_msg += f': {repr(e)}'
        logger.error(error_msg)
        messages.error(request, error_msg)
    return None
